/**
  CC License 2020 by ACME CAD CAM
  3D additive printer post configuration.

  $Revision: 42614 ccbc0b14704b013ada243633a5bc25f45bcb39f5 $
  $Date: 2019-12-20 12:00:51 $

  FORKID {A316FBC4-FA6E-41C5-A347-3D94F72F5D06}
*/

description = 'Reprap and Duet3d Firmware'
vendor = 'ACME CAD CAM'
vendorUrl = 'http://acmecadcam.com'
legal = 'Attribution-NonCommercial-ShareAlike 4.0 International'
certificationLevel = 2
minimumRevision = 45633

longDescription =
  'Simple post to export toolpath for Duet3D machines in reprap firmware gcode format'

extension = 'gcode'
setCodePage('ascii')

capabilities = CAPABILITY_ADDITIVE
tolerance = spatial(0.002, MM)
highFeedrate = unit == MM ? 6000 : 236

// needed for range checking, will be effectively passed from Fusion
var printerLimits = {
  x: { min: 0, max: 300.0 }, //Defines the x bed size
  y: { min: 0, max: 300.0 }, //Defines the y bed size
  z: { min: 0, max: 300.0 } //Defines the z bed size
}

// User-defined properties
properties = {
  stanbyTemp: 0,
  toolOverride: 0
}

// User-defined property definitions
propertyDefinitions = {
  stanbyTemp: {
    title: 'Standby Temp',
    description: 'Specifies the standby temperature for extruders',
    type: 'number'
  },
  toolOverride: {
    title: 'Tool 0 Override',
    description: 'Override the priamry tool with a specific ( Set tool 0 - 3)',
    type: 'number'
  }
}

// Workaround properties
var extruderOffsets = [
  [0, 0, 0],
  [0, 0, 0]
]
var activeExtruder = 0 // Track the active extruder

var totalFilament = 0 // Track the total fillament

var zHolder = 0
var layerOneHeight = 0
var layerTwoHeight = 0
var layerAllHeight = 0

var xyzFormat = createFormat({ decimals: unit == MM ? 3 : 4 })
var xFormat = createFormat({ decimals: unit == MM ? 3 : 4 })
var yFormat = createFormat({ decimals: unit == MM ? 3 : 4 })
var zFormat = createFormat({ decimals: unit == MM ? 3 : 4 })
var gFormat = createFormat({
  prefix: 'G',
  width: 1,
  zeropad: false,
  decimals: 0
})
var mFormat = createFormat({
  prefix: 'M',
  width: 2,
  zeropad: true,
  decimals: 0
})
var tFormat = createFormat({
  prefix: 'T',
  width: 1,
  zeropad: false,
  decimals: 0
})
var pFormat = createFormat({
  prefix: 'P',
  zeropad: false,
  decimals: 0
})
var feedFormat = createFormat({ decimals: unit == MM ? 0 : 1 })
var heightFormat = createFormat({ decimals: 2 })
var integerFormat = createFormat({ decimals: 0 })
var dimensionFormat = createFormat({
  decimals: unit == MM ? 3 : 4,
  zeropad: false,
  suffix: unit == MM ? 'mm' : 'in'
})

var gMotionModal = createModal(
  {
    force: true
  },
  gFormat
) // Modal group 1 // G0-G3, ...
var gPlaneModal = createModal(
  {
    onchange: function () {
      gMotionModal.reset()
    }
  },
  gFormat
) // Modal group 2 // G17-19 //Actually unused
var gAbsIncModal = createModal({}, gFormat) // Modal group 3 // G90-91

var xOutput = createVariable({ prefix: 'X' }, xFormat)
var yOutput = createVariable({ prefix: 'Y' }, yFormat)
var zOutput = createVariable({ prefix: 'Z' }, zFormat)
var feedOutput = createVariable({ prefix: 'F' }, feedFormat)
var eOutput = createVariable({ prefix: 'E' }, xyzFormat) // Extrusion length
var sOutput = createVariable({ prefix: 'S', force: true }, xyzFormat) // Parameter temperature or speed
var rOutput = createVariable({ prefix: 'R', force: true }, xyzFormat)

// Start G-codes
function onOpen () {
  getPrinterGeometry()

  if (programName) {
    writeComment(programName)
  }
  if (programComment) {
    writeComment(programComment)
  }

  var pad = function (num, size) {
      return ('000' + num).slice(size * -1)
    },
    // Time calculation
    time = parseFloat(printTime).toFixed(3),
    hours = Math.floor(time / 60 / 60),
    minutes = Math.floor(time / 60) % 60,
    seconds = Math.floor(time - minutes * 60),
    pTime =
      pad(hours, 2) +
      ' hours ' +
      pad(minutes, 2) +
      ' minutes ' +
      pad(seconds, 2) +
      ' seconds'

  // Write Job info
  writeComment('Build time: ' + pTime)
  writeComment('G-code generated on: ' + getGlobalParameter('generated-at'))
  writeComment(
    'G-code generated by Fusion 360 ' + getGlobalParameter('version')
  )
  writeComment('Username: ' + getGlobalParameter('username'))
  writeComment('Layer count: ' + getGlobalParameter('layer-cnt'))
  writeComment('Number of bodies: ' + integerFormat.format(partCount))
  writeComment(
    'Extruder 1 Material used: ' +
      dimensionFormat.format(getExtruder(1).extrusionLength)
  )

  totalFilament = totalFilament + getExtruder(1).extrusionLength

  writeComment('Extruder 1 Material name: ' + getExtruder(1).materialName)
  writeComment(
    'Extruder 1 Filament diameter: ' +
      dimensionFormat.format(getExtruder(1).filamentDiameter)
  )
  writeComment(
    'Extruder 1 Nozzle diameter: ' +
      dimensionFormat.format(getExtruder(1).nozzleDiameter)
  )
  writeComment(
    'Extruder 1 offset x: ' + dimensionFormat.format(extruderOffsets[0][0])
  )
  writeComment(
    'Extruder 1 offset y: ' + dimensionFormat.format(extruderOffsets[0][1])
  )
  writeComment(
    'Extruder 1 offset z: ' + dimensionFormat.format(extruderOffsets[0][2])
  )

  if (
    hasGlobalParameter('ext2-extrusion-len') &&
    hasGlobalParameter('ext2-nozzle-dia') &&
    hasGlobalParameter('ext2-temp') &&
    hasGlobalParameter('ext2-filament-dia') &&
    hasGlobalParameter('ext2-material-name')
  ) {
    writeComment(
      'Extruder 2 material used: ' +
        dimensionFormat.format(getExtruder(2).extrusionLength)
    )

    totalFilament = totalFilament + getExtruder(2).extrusionLength

    writeComment('Extruder 2 material name: ' + getExtruder(2).materialName)
    writeComment(
      'Extruder 2 filament diameter: ' +
        dimensionFormat.format(getExtruder(2).filamentDiameter)
    )
    writeComment(
      'Extruder 2 nozzle diameter: ' +
        dimensionFormat.format(getExtruder(2).nozzleDiameter)
    )
    writeComment(
      'Extruder 2 max temp: ' + integerFormat.format(getExtruder(2).temperature)
    )
    writeComment(
      'Extruder 2 offset x: ' + dimensionFormat.format(extruderOffsets[1][0])
    )
    writeComment(
      'Extruder 2 offset y: ' + dimensionFormat.format(extruderOffsets[1][1])
    )
    writeComment(
      'Extruder 2 offset z: ' + dimensionFormat.format(extruderOffsets[1][2])
    )
  }
  writeComment('Max temp: ' + integerFormat.format(getExtruder(1).temperature))
  writeComment('Bed temp: ' + integerFormat.format(bedTemp))
  writeComment('Standby temp; ' + properties.stanbyTemp)
  writeComment('Print volume X: ' + dimensionFormat.format(printerLimits.x.max))
  writeComment('Print volume Y: ' + dimensionFormat.format(printerLimits.y.max))
  writeComment('Print volume Z: ' + dimensionFormat.format(printerLimits.z.max))
}

function onSection () {
  var range = currentSection.getBoundingBox()
  axes = ['x', 'y', 'z']
  formats = [xFormat, yFormat, zFormat]
  for (var element in axes) {
    var min = formats[element].getResultingValue(range.lower[axes[element]])
    var max = formats[element].getResultingValue(range.upper[axes[element]])
    if (
      printerLimits[axes[element]].max < max ||
      printerLimits[axes[element]].min > min
    ) {
      error(localize('A toolpath is outside of the build volume.'))
    }
  }

  // Set unit
  switch (unit) {
    case IN:
      writeBlock(gFormat.format(20) + ' ; Use inches')
      break
    case MM:
      writeBlock(gFormat.format(21) + ' ; Use mm')
      break
  }
  writeBlock(gAbsIncModal.format(90)) // Absolute spatial co-ordinates
  writeBlock(mFormat.format(82)) // Absolute extrusion co-ordinates
}

// End G-codes
function onClose () {
  writeBlock(tFormat.format(-1) + ' ; Drop tool off')
  writeBlock(mFormat.format(400) + ' ; Clear move buffer')
  writeBlock(mFormat.format(117) + ' PRINT FINISHED')
  writeBlock('M0 ; All heaters off')
  writeComment('END OF GCODE')
  writeComment('--------------------------------')
  writeComment('Print Statistics')
  writeComment('--------------------------------')
  writeComment('Fist Layer height: ' + layerOneHeight)
  writeComment('Layer height: ' + layerAllHeight)
  writeComment('Layer count: ' + getGlobalParameter('layer-cnt'))
  writeComment('Filament length: ' + dimensionFormat.format(totalFilament))
}

// Writes the specified block.
function writeBlock () {
  writeWords(arguments)
}

function getPrinterGeometry () {
  machineConfiguration = getMachineConfiguration()

  // Get the printer geometry from the machine configuration
  printerLimits.x.min = 0 - machineConfiguration.getCenterPositionX()
  printerLimits.y.min = 0 - machineConfiguration.getCenterPositionY()
  printerLimits.z.min = 0 + machineConfiguration.getCenterPositionZ()

  printerLimits.x.max =
    machineConfiguration.getWidth() - machineConfiguration.getCenterPositionX()
  printerLimits.y.max =
    machineConfiguration.getDepth() - machineConfiguration.getCenterPositionY()
  printerLimits.z.max =
    machineConfiguration.getHeight() + machineConfiguration.getCenterPositionZ()

  extruderOffsets[0][0] = machineConfiguration.getExtruderOffsetX(1)
  extruderOffsets[0][1] = machineConfiguration.getExtruderOffsetY(1)
  extruderOffsets[0][2] = machineConfiguration.getExtruderOffsetZ(1)
  if (numberOfExtruders > 1) {
    extruderOffsets[1] = []
    extruderOffsets[1][0] = machineConfiguration.getExtruderOffsetX(2)
    extruderOffsets[1][1] = machineConfiguration.getExtruderOffsetY(2)
    extruderOffsets[1][2] = machineConfiguration.getExtruderOffsetZ(2)
  }
}

// Write comments
function onComment (message) {
  writeComment(message)
}

function onRapid (_x, _y, _z) {
  var x = xOutput.format(_x)
  var y = yOutput.format(_y)
  var z = zOutput.format(_z)
  zHolder = _z
  if (x || y || z) {
    writeBlock(gMotionModal.format(0), x, y, z)
  }
}

function onLinearExtrude (_x, _y, _z, _f, _e) {
  var x = xOutput.format(_x)
  var y = yOutput.format(_y)
  var z = zOutput.format(_z)
  var f = feedOutput.format(_f)
  var e = eOutput.format(_e)
  if (x || y || z || f || e) {
    writeBlock(gMotionModal.format(1), x, y, z, f, e)
  }
}

function onBedTemp (temp, wait) {
  if (wait) {
    writeBlock(mFormat.format(190), sOutput.format(temp) + ' ; DELETE ME')
  } else {
    writeBlock(mFormat.format(140), sOutput.format(temp))
  }
}

function onExtruderChange (id) {
  if (id < numberOfExtruders) {
    writeBlock(tFormat.format(id) + ' ; Us tool ' + id)
    activeExtruder = id
    xOutput.reset()
    yOutput.reset()
    zOutput.reset()
    writeBlock(gFormat.format(29) + ' S1')
  } else {
    error(
      localize("This printer doesn't support the extruder ") +
        integerFormat.format(id) +
        ' !'
    )
  }
}

function onExtrusionReset (length) {
  eOutput.reset()
  writeBlock(gFormat.format(92), eOutput.format(length))
}

function onLayer (num) {
  if (num == 2) {
    layerOneHeight = heightFormat.format(zHolder)
  }

  if (num == 3) {
    layerTwoHeight = heightFormat.format(zHolder)
    layerAllHeight = layerTwoHeight - layerOneHeight
  }

  writeComment(
    'Layer : ' +
      integerFormat.format(num) +
      ' of ' +
      integerFormat.format(layerCount)
  )
}

function onExtruderTemp (temp, wait, id) {
  if (id < numberOfExtruders) {
    if (id == 0) {
      id = properties.toolOverride
      if (wait) {
        writeBlock(
          gFormat.format(10),
          pFormat.format(id),
          sOutput.format(temp),
          rOutput.format(properties.stanbyTemp)
        )
        writeBlock(tFormat.format(id) + ' ; Use Tool ' + id)
        writeBlock(gFormat.format(29) + ' S1')
        writeBlock(mFormat.format(116))
      } else {
        writeBlock(
          gFormat.format(10),
          pFormat.format(id),
          sOutput.format(temp),
          rOutput.format(properties.stanbyTemp) + ' ; DELETE ME'
        )
      }
    } else {
    }
  } else {
    error(
      localize("This printer doesn't support the extruder ") +
        integerFormat.format(id) +
        ' !'
    )
  }
}

function onFanSpeed (speed, id) {
  // TODO handle id information
  if (speed == 0) {
    writeBlock(mFormat.format(107) + ' ; Fan off')
  } else {
    writeBlock(mFormat.format(106), sOutput.format(speed) + ' ; Set fan')
  }
}

function onParameter (name, value) {
  switch (name) {
    // Feedrate is set before rapid moves and extruder change
    case 'feedRate':
      setFeedRate(value)
      break
    case 'customCommand':
      if (value == 'start_gcode') {
        // Anyhting you want to write before setting temps
        writeBlock(gFormat.format(28) + ' Z ; Probe Z')
      }
      if (value == 'end_gcode') {
        // Anyhting you want to write before end gcodes
      }
      break
    // Warning or error message on unhandled parameter?
  }
}

// User defined functions
function setFeedRate (value) {
  feedOutput.reset()
  writeBlock(gFormat.format(1), feedOutput.format(value))
}

// Write comments
function writeComment (text) {
  writeln('; ' + text)
}

// Write G-code pass through custom commands
function writeCustomCommand (text) {
  if (text.length > 0) {
    writeln(text)
  }
}
